'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from enum import Enum
from skyflow._utils import log_error


class SkyflowErrorCodes(Enum):
    INVALID_INPUT = 400
    SERVER_ERROR = 500
    PARTIAL_SUCCESS = 500


class SkyflowErrorMessages(Enum):
    API_ERROR = "Server returned status code %s"

    FILE_NOT_FOUND = "File at %s not found"
    FILE_INVALID_JSON = "File at %s is not in JSON format"
    INVALID_CREDENTIALS = "Given credentials are not valid"
    INVALID_URL = "Given url '%s' is invalid"

    MISSING_PRIVATE_KEY = "Unable to read Private key"
    MISSING_CLIENT_ID = "Unable to read Client ID"
    MISSING_KEY_ID = "Unable to read Key ID"
    MISSING_TOKEN_URI = "Unable to read Token URI"

    JWT_INVALID_FORMAT = "Private key is not in correct format"
    MISSING_ACCESS_TOKEN = "accessToken not present in response"
    MISSING_TOKEN_TYPE = "tokenType not present in response"
    JWT_DECODE_ERROR = "Invalid access token"

    # vault
    RECORDS_KEY_ERROR = "Records key is missing from payload"
    FIELDS_KEY_ERROR = "Fields key is missing from payload"
    TABLE_KEY_ERROR = "Table key is missing from payload"
    TOKEN_KEY_ERROR = "Token key is missing from payload"
    IDS_KEY_ERROR = "Ids key is missing from payload"
    REDACTION_KEY_ERROR = "Redaction key is missing from payload"

    INVALID_JSON = "Given %s is invalid JSON"
    INVALID_RECORDS_TYPE = "Records key has value of type %s, expected list"
    INVALID_FIELDS_TYPE = "Fields key has value of type %s, expected string"
    INVALID_TABLE_TYPE = "Table key has value of type %s, expected string"
    INVALID_IDS_TYPE = "Ids key has value of type %s, expected list"
    INVALID_ID_TYPE = "Id key has value of type %s, expected string"
    INVALID_REDACTION_TYPE = "Redaction key has value of type %s, expected Skyflow.Redaction"

    INVALID_REQUEST_BODY = "Given request body is not valid"
    INVALID_RESPONSE_BODY = "Given response body is not valid"
    INVALID_HEADERS = "Given Request Headers is not valid"
    INVALID_PATH_PARAMS = "Given path params are not valid"
    INVALID_QUERY_PARAMS = "Given query params are not valid"
    INVALID_PATH_PARAM_TYPE = "Path params (key, value) must be of type 'str' given type - (%s, %s)"
    INVALID_QUERY_PARAM_TYPE = "Query params (key, value) must be of type 'str' given type - (%s, %s)"

    INVALID_TOKEN_TYPE = "Token key has value of type %s, expected string"
    PARTIAL_SUCCESS = "Server returned errors, check SkyflowError.data for more"

    VAULT_ID_INVALID_TYPE = "Expected Vault ID to be str, got %s"
    VAULT_URL_INVALID_TYPE = "Expected Vault URL to be str, got %s"
    TOKEN_PROVIDER_ERROR = "Expected Token Provider to be function, got %s"

    EMPTY_VAULT_ID = "Vault ID must not be empty"
    EMPTY_VAULT_URL = "Vault URL must not be empty"
    RESPONSE_NOT_JSON = "Response %s is not valid JSON"

    TOKEN_PROVIDER_INVALID_TOKEN = "Invalid token from tokenProvider"
    INVALID_UPSERT_OPTIONS_TYPE = "upsertOptions key has value of type %s, expected list"
    EMPTY_UPSERT_OPTIONS_LIST = "upsert option cannot be an empty array, atleast one object of table and column is required"
    INVALID_UPSERT_TABLE_TYPE = "upsert object table key has value of type %s, expected string"
    INVALID_UPSERT_COLUMN_TYPE = "upsert object column key has value of type %s, expected string"
    EMPTY_UPSERT_OPTION_TABLE = "upsert object table value is empty string at index %s, expected non-empty string"
    EMPTY_UPSERT_OPTION_COLUMN = "upsert object column value is empty string at index %s, expected non-empty string"

class SkyflowError(Exception):
    def __init__(self, code, message="An Error occured", data={}, interface: str = 'Unknown') -> None:
        if type(code) is SkyflowErrorCodes:
            self.code = code.value
        else:
            self.code = code
        if type(message) is SkyflowErrorMessages:
            self.message = message.value
        else:
            self.message = message
        log_error(self.message, interface)
        self.data = data
        super().__init__(self.message)



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from skyflow.errors._skyflow_errors import SkyflowError, SkyflowErrorCodes, SkyflowErrorMessages
import asyncio
from aiohttp import ClientSession, request
import json
from skyflow._utils import InterfaceName

interface = InterfaceName.DETOKENIZE.value


def getDetokenizeRequestBody(data):
    try:
        token = data["token"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.TOKEN_KEY_ERROR, interface=interface)
    if not isinstance(token, str):
        tokenType = str(type(token))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_TOKEN_TYPE.value % (
            tokenType), interface=interface)
    requestBody = {"detokenizationParameters": []}
    requestBody["detokenizationParameters"].append({
        "token": token})
    return requestBody


async def sendDetokenizeRequests(data, url, token):

    tasks = []

    try:
        records = data["records"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.RECORDS_KEY_ERROR, interface=interface)
    if not isinstance(records, list):
        recordsType = str(type(records))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_RECORDS_TYPE.value % (
            recordsType), interface=interface)

    validatedRecords = []
    for record in records:
        requestBody = getDetokenizeRequestBody(record)
        jsonBody = json.dumps(requestBody)
        validatedRecords.append(jsonBody)
    async with ClientSession() as session:
        for record in validatedRecords:
            headers = {
                "Authorization": "Bearer " + token
            }
            task = asyncio.ensure_future(post(url, record, headers, session))
            tasks.append(task)
        await asyncio.gather(*tasks)
        await session.close()
    return tasks


async def post(url, data, headers, session):
    async with session.post(url, data=data, headers=headers, ssl=False) as response:
        try:
            return (await response.read(), response.status, response.headers['x-request-id'])
        except KeyError:
            return (await response.read(), response.status)


def createDetokenizeResponseBody(responses):
    result = {
        "records": [],
        "errors": []
    }
    partial = False
    for response in responses:
        r = response.result()
        status = r[1]
        try:
            jsonRes = json.loads(r[0].decode('utf-8'))
        except:
            raise SkyflowError(status,
                               SkyflowErrorMessages.RESPONSE_NOT_JSON.value % r[0].decode('utf-8'), interface=interface)

        if status == 200:
            temp = {}
            temp["token"] = jsonRes["records"][0]["token"]
            temp["value"] = jsonRes["records"][0]["value"]
            result["records"].append(temp)
        else:
            temp = {"error": {}}
            temp["error"]["code"] = jsonRes["error"]["http_code"]
            temp["error"]["description"] = jsonRes["error"]["message"]
            if len(r) > 2 and r[2] != None:
                temp["error"]["description"] += ' - Request ID: ' + str(r[2])
            result["errors"].append(temp)
            partial = True
    return result, partial



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
import json
import jwt
import datetime
import requests
from warnings import warn
from collections import namedtuple
from skyflow._utils import log_info, InterfaceName, InfoMessages


from skyflow.errors._skyflow_errors import *

ResponseToken = namedtuple('ResponseToken', ['AccessToken', 'TokenType'])
interface = InterfaceName.GENERATE_BEARER_TOKEN


def generate_bearer_token(credentialsFilePath: str) -> ResponseToken:
    '''
    This function is used to get the access token for skyflow Service Accounts.
    `credentialsFilePath` is the file path in string of the credentials file that is downloaded after Service Account creation.

    Response Token is a named tupe with two attributes:
        1. AccessToken: The access token
        2. TokenType: The type of access token (eg: Bearer)
    '''

    log_info(InfoMessages.GENERATE_BEARER_TOKEN_TRIGGERED.value,
             interface=interface)

    try:
        credentialsFile = open(credentialsFilePath, 'r')
    except:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.FILE_NOT_FOUND.value % (credentialsFilePath), interface=interface)

    try:
        credentials = json.load(credentialsFile)
    except Exception:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.FILE_INVALID_JSON.value % (credentialsFilePath), interface=interface)
    finally:
        credentialsFile.close()

    result = getSAToken(credentials)

    log_info(InfoMessages.GENERATE_BEARER_TOKEN_SUCCESS.value,
             interface=interface)
    return result


def generate_bearer_token_from_creds(credentials: str) -> ResponseToken:
    '''
    This function is used to get the access token for skyflow Service Accounts.
    `credentials` arg takes the content of the credentials file that is downloaded after Service Account creation.

    Response Token is a named tupe with two attributes:
        1. AccessToken: The access token
        2. TokenType: The type of access token (eg: Bearer)
    '''

    log_info(InfoMessages.GENERATE_BEARER_TOKEN_TRIGGERED.value,
             interface=interface)
    try:
        jsonCredentials = json.loads(credentials)
    except Exception as e:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_CREDENTIALS, interface=interface)
    result = getSAToken(jsonCredentials)

    log_info(InfoMessages.GENERATE_BEARER_TOKEN_SUCCESS.value,
             interface=interface)
    return result


def getSAToken(credentials):
    try:
        privateKey = credentials["privateKey"]
    except:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.MISSING_PRIVATE_KEY, interface=interface)
    try:
        clientID = credentials["clientID"]
    except:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.MISSING_CLIENT_ID, interface=interface)
    try:
        keyID = credentials["keyID"]
    except:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.MISSING_KEY_ID, interface=interface)
    try:
        tokenURI = credentials["tokenURI"]
    except:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.MISSING_TOKEN_URI, interface=interface)

    signedToken = getSignedJWT(clientID, keyID, tokenURI, privateKey)

    response = sendRequestWithToken(tokenURI, signedToken)
    content = response.content.decode('utf-8')

    try:
        token = json.loads(content)
    except json.decoder.JSONDecodeError as e:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.RESPONSE_NOT_JSON % content, interface=interface)
    return getResponseToken(token)


def getSignedJWT(clientID, keyID, tokenURI, privateKey):
    payload = {
        "iss": clientID,
        "key": keyID,
        "aud": tokenURI,
        "sub": clientID,
        "exp": datetime.datetime.utcnow() + datetime.timedelta(minutes=60)
    }
    try:
        return jwt.encode(payload=payload, key=privateKey, algorithm="RS256")
    except Exception as e:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.JWT_INVALID_FORMAT, interface=interface)


def sendRequestWithToken(url, token):
    headers = {
        "content-type": "application/json"
    }
    payload = {
        "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
        "assertion":  token
    }
    try:
        response = requests.post(url=url, json=payload, headers=headers)
        statusCode = response.status_code
    except requests.exceptions.InvalidURL:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_URL.value % (url), interface=interface)
    except requests.exceptions.MissingSchema:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_URL.value % (url), interface=interface)

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as error:
        message = SkyflowErrorMessages.API_ERROR.value % statusCode
        if error.response != None and error.response.content != None:
            try:
                errorResponse = json.loads(
                    error.response.content.decode('utf-8'))
                if 'error' in errorResponse and type(errorResponse['error']) == type({}) and 'message' in errorResponse['error']:
                    message = errorResponse['error']['message']
            except:
                message = SkyflowErrorMessages.RESPONSE_NOT_JSON.value % error.response.content.decode(
                    'utf-8')
        if 'x-request-id' in response.headers:
            message += ' - request id: ' + response.headers['x-request-id']
        raise SkyflowError(statusCode, message, interface=interface)

    return response


def getResponseToken(token):
    try:
        accessToken = token["accessToken"]
    except:
        raise SkyflowError(SkyflowErrorCodes.SERVER_ERROR,
                           SkyflowErrorMessages.MISSING_ACCESS_TOKEN, interface=interface)

    try:
        tokenType = token["tokenType"]
    except:
        raise SkyflowError(SkyflowErrorCodes.SERVER_ERROR,
                           SkyflowErrorMessages.MISSING_TOKEN_TYPE, interface=interface)

    return ResponseToken(AccessToken=accessToken, TokenType=tokenType)



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from warnings import warn
import jwt
import time
from skyflow.errors._skyflow_errors import *
from skyflow._utils import InterfaceName, log_info, log_error, InfoMessages


def is_expired(token: str):
    '''
    Check if stored token is not expired, if not return a new token, 
    if the token has expiry time before 5min of current time, call returns False
    '''
    interface = InterfaceName.IS_EXPIRED.value
    log_info(InfoMessages.IS_EXPIRED_TRIGGERED.value, interface=interface)
    if len(token) == 0:
        log_info(InfoMessages.EMPTY_ACCESS_TOKEN, interface=interface)
        return True

    try:
        decoded = jwt.decode(
            token, options={"verify_signature": False, "verify_aud": False})
        if time.time() < decoded['exp']:
            return False
    except jwt.ExpiredSignatureError:
        return True
    except Exception as e:
        log_error(InfoMessages.INVALID_TOKEN.value, interface=interface)
        return True

    return True



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
import urllib.parse
import logging
from enum import Enum

skyflowLog = logging.getLogger('skyflow')
skyflowLog.setLevel(logging.ERROR)

supported_content_types = {
    "JSON": 'application/json',
    "PLAINTEXT": 'text/plain',
    "XML": 'text/xml',
    "URLENCODED": 'application/x-www-form-urlencoded',
    "FORMDATA": 'multipart/form-data',
}


class LogLevel(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARN
    ERROR = logging.ERROR
    OFF = logging.CRITICAL


def set_log_level(logLevel: LogLevel):
    '''
    Sets the Log Level for the Skyflow python SDK
    '''
    skyflowLog.setLevel(logLevel.value)


def log_info(message: str, interface: str):
    formattedMessage = '{} {}'.format(interface, message)
    skyflowLog.info(formattedMessage)


# def log_debug(message: str, interface: str):
#     formattedMessage = '{} {}'.format(interface, message)
#     skyflowLog.debug(formattedMessage)


def log_error(message: str, interface: str):
    formattedMessage = '{} {}'.format(interface, message)
    skyflowLog.error(formattedMessage)


class InfoMessages(Enum):
    INITIALIZE_CLIENT = "Initializing skyflow client"
    CLIENT_INITIALIZED = "Initialized skyflow client successfully"
    VALIDATE_INSERT_RECORDS = "Validating insert records"
    VALIDATE_DETOKENIZE_INPUT = "Validating detokenize input"
    VALIDATE_GET_BY_ID_INPUT = "Validating getByID input"
    VALIDATE_CONNECTION_CONFIG = "Validating connection config"
    INSERT_DATA_SUCCESS = "Data has been inserted successfully."
    DETOKENIZE_SUCCESS = "Data has been detokenized successfully."
    GET_BY_ID_SUCCESS = "Data fetched from ID successfully."
    BEARER_TOKEN_RECEIVED = "tokenProvider returned token successfully."
    INSERT_TRIGGERED = "Insert method triggered."
    DETOKENIZE_TRIGGERED = "Detokenize method triggered."
    GET_BY_ID_TRIGGERED = "Get by ID triggered."
    INVOKE_CONNECTION_TRIGGERED = "Invoke connection triggered."
    GENERATE_BEARER_TOKEN_TRIGGERED = "Generate bearer token triggered"
    GENERATE_BEARER_TOKEN_SUCCESS = "Generate bearer token returned successfully"
    IS_TOKEN_VALID_TRIGGERED = "isTokenValid() triggered"
    IS_EXPIRED_TRIGGERED = "is_expired() triggered"
    EMPTY_ACCESS_TOKEN = "Give access token is empty"
    INVALID_TOKEN = "Given token is invalid"


class InterfaceName(Enum):
    CLIENT = "client"
    INSERT = "client.insert"
    DETOKENIZE = "client.detokenize"
    GET_BY_ID = "client.get_by_id"
    INVOKE_CONNECTION = "client.invoke_connection"
    GENERATE_BEARER_TOKEN = "service_account.generate_bearer_token"

    IS_TOKEN_VALID = "service_account.isTokenValid"
    IS_EXPIRED = "service_account.is_expired"


def http_build_query(data):
    '''
        Creates a form urlencoded string from python dictionary
        urllib.urlencode() doesn't encode it in a php-esque way, this function helps in that
    '''

    return urllib.parse.urlencode(r_urlencode(list(), dict(), data))


def r_urlencode(parents, pairs, data):
    '''
        convert the python dict recursively into a php style associative dictionary
    '''
    if isinstance(data, list) or isinstance(data, tuple):
        for i in range(len(data)):
            parents.append(i)
            r_urlencode(parents, pairs, data[i])
            parents.pop()
    elif isinstance(data, dict):
        for key, value in data.items():
            parents.append(key)
            r_urlencode(parents, pairs, value)
            parents.pop()
    else:
        pairs[render_key(parents)] = str(data)

    return pairs


def render_key(parents):
    '''
        renders the nested dictionary key as an associative array (php style dict)
    '''
    depth, outStr = 0, ''
    for x in parents:
        s = "[%s]" if depth > 0 or isinstance(x, int) else "%s"
        outStr += s % str(x)
        depth += 1
    return outStr



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
import jwt
import time
from skyflow.errors._skyflow_errors import *


def tokenProviderWrapper(storedToken: str, newTokenProvider, interface: str):
    '''
    Check if stored token is not expired, if not return a new token
    '''

    if len(storedToken) == 0:
        newToken = newTokenProvider()
        verify_token_from_provider(newToken, interface)
        return newToken

    try:
        decoded = jwt.decode(storedToken, options={
                             "verify_signature": False, "verify_aud": False})
        if time.time() + 300 < decoded['exp']:
            return storedToken
        else:
            newToken = newTokenProvider()
            verify_token_from_provider(newToken, interface)
            return newToken
    except Exception:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.JWT_DECODE_ERROR, interface=interface)


def verify_token_from_provider(token, interface):
    '''
    Verify the jwt from token provider
    '''
    try:
        jwt.decode(token, options={
            "verify_signature": False,
            "verify_aud": False
        }, algorithms=['RS256'])
    except Exception as e:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.TOKEN_PROVIDER_INVALID_TOKEN, interface=interface)



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from requests.sessions import PreparedRequest
from ._config import ConnectionConfig
from skyflow.errors._skyflow_errors import *
import requests
import json

from skyflow._utils import InterfaceName, http_build_query, supported_content_types, r_urlencode

interface = InterfaceName.INVOKE_CONNECTION.value


def createRequest(config: ConnectionConfig) -> PreparedRequest:
    url = parsePathParams(config.connectionURL.rstrip('/'), config.pathParams)

    try:
        if isinstance(config.requestHeader, dict):
            header = to_lowercase_keys(json.loads(
                json.dumps(config.requestHeader)))
        else:
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                               SkyflowErrorMessages.INVALID_REQUEST_BODY, interface=interface)
    except Exception:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_HEADERS, interface=interface)
    if not 'Content-Type'.lower() in header:
        header['content-type'] = supported_content_types["JSON"]

    try:
        if isinstance(config.requestBody, dict):
            json_data, files = get_data_from_content_type(
                config.requestBody, header["content-type"])
        else:
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                               SkyflowErrorMessages.INVALID_RESPONSE_BODY, interface=interface)
    except Exception as e:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_REQUEST_BODY, interface=interface)

    verifyParams(config.queryParams, config.pathParams)

    try:
        return requests.Request(
            method=config.methodName.value,
            url=url,
            data=json_data,
            headers=header,
            params=config.queryParams,
            files=files
        ).prepare()
    except requests.exceptions.InvalidURL:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_URL.value % (
            config.connectionURL), interface=interface)


def parsePathParams(url, pathParams):
    result = url
    for param, value in pathParams.items():
        result = result.replace('{' + param + '}', value)

    return result


def verifyParams(queryParams, pathParams):
    if not isinstance(pathParams, dict):
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_PATH_PARAMS, interface=interface)
    if not isinstance(queryParams, dict):
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_QUERY_PARAMS, interface=interface)

    for param, value in pathParams.items():
        if not(isinstance(param, str) and isinstance(value, str)):
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_PATH_PARAM_TYPE.value % (
                str(type(param)), str(type(value))), interface=interface)

    for param, value in queryParams.items():
        if not isinstance(param, str):
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_QUERY_PARAM_TYPE.value % (
                str(type(param)), str(type(value))), interface=interface)

    try:
        json.dumps(queryParams)
    except TypeError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_QUERY_PARAMS, interface=interface)


def to_lowercase_keys(dict):
    '''
        convert keys of dictionary to lowercase
    '''
    result = {}
    for key, value in dict.items():
        result[key.lower()] = value

    return result


def get_data_from_content_type(data, content_type):
    '''
        Get request data according to content type
    '''
    converted_data = data
    files = {}
    if content_type == supported_content_types["URLENCODED"]:
        converted_data = http_build_query(data)
    elif content_type == supported_content_types["FORMDATA"]:
        converted_data = r_urlencode(list(), dict(), data)
        files = {(None, None)}
    elif content_type == supported_content_types["JSON"]:
        converted_data = json.dumps(data)

    return converted_data, files



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from ._token import generate_bearer_token
from ._token import generate_bearer_token
from ._token import ResponseToken
from ._token import generate_bearer_token_from_creds
from ._validity import is_expired



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from skyflow.errors._skyflow_errors import SkyflowError, SkyflowErrorCodes, SkyflowErrorMessages
import asyncio
from aiohttp import ClientSession
import json
from ._config import RedactionType
from skyflow._utils import InterfaceName

interface = InterfaceName.GET_BY_ID.value


def getGetByIdRequestBody(data):
    try:
        ids = data["ids"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.IDS_KEY_ERROR, interface=interface)
    if not isinstance(ids, list):
        idsType = str(type(ids))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.INVALID_IDS_TYPE.value % (idsType), interface=interface)
    for id in ids:
        if not isinstance(id, str):
            idType = str(type(id))
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_ID_TYPE.value % (
                idType), interface=interface)
    try:
        table = data["table"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.TABLE_KEY_ERROR, interface=interface)
    if not isinstance(table, str):
        tableType = str(type(table))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_TABLE_TYPE.value % (
            tableType), interface=interface)
    try:
        redaction = data["redaction"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.REDACTION_KEY_ERROR, interface=interface)
    if not isinstance(redaction, RedactionType):
        redactionType = str(type(redaction))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_REDACTION_TYPE.value % (
            redactionType), interface=interface)
    return ids, table, redaction.value


async def sendGetByIdRequests(data, url, token):
    tasks = []
    try:
        records = data["records"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.RECORDS_KEY_ERROR, interface=interface)
    if not isinstance(records, list):
        recordsType = str(type(records))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_RECORDS_TYPE.value % (
            recordsType), interface=interface)

    validatedRecords = []
    for record in records:
        ids, table, redaction = getGetByIdRequestBody(record)
        validatedRecords.append((ids, table, redaction))
    async with ClientSession() as session:
        for record in validatedRecords:
            headers = {
                "Authorization": "Bearer " + token
            }
            params = {"skyflow_ids": record[0], "redaction": record[2]}
            task = asyncio.ensure_future(
                get(url, headers, params, session, record[1]))
            tasks.append(task)
        await asyncio.gather(*tasks)
        await session.close()
    return tasks


async def get(url, headers, params, session, table):
    async with session.get(url + "/" + table, headers=headers, params=params, ssl=False) as response:
        try:
            return (await response.read(), response.status, table, response.headers['x-request-id'])
        except KeyError:
            return (await response.read(), response.status, table)


def createGetByIdResponseBody(responses):
    result = {
        "records": [],
        "errors": []
    }
    for response in responses:
        partial = False
        r = response.result()
        status = r[1]
        try:
            jsonRes = json.loads(r[0].decode('utf-8'))
        except:
            raise SkyflowError(status,
                               SkyflowErrorMessages.RESPONSE_NOT_JSON.value % r[0].decode('utf-8'), interface=interface)

        if status == 200:
            changedRecords = []
            for record in jsonRes["records"]:
                temp = record
                temp["table"] = r[2]
                changedRecords.append(temp)
            result["records"] += changedRecords
        else:
            temp = {"error": {}}
            temp["error"]["code"] = jsonRes["error"]["http_code"]
            temp["error"]["description"] = jsonRes["error"]["message"]
            if len(r) > 3 and r[3] != None:
                temp["error"]["description"] += ' - Request ID: ' + str(r[3])
            result["errors"].append(temp)
            result["errors"].append(temp)
            partial = True
    return result, partial



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
import types
import requests
from ._insert import getInsertRequestBody, processResponse, convertResponse
from ._config import Configuration
from ._config import InsertOptions, ConnectionConfig
from ._connection import createRequest
from ._detokenize import sendDetokenizeRequests, createDetokenizeResponseBody
from ._get_by_id import sendGetByIdRequests, createGetByIdResponseBody
import asyncio
from skyflow.errors._skyflow_errors import SkyflowError, SkyflowErrorCodes, SkyflowErrorMessages
from skyflow._utils import log_info, InfoMessages, InterfaceName
from ._token import tokenProviderWrapper


class Client:
    def __init__(self, config: Configuration):

        interface = InterfaceName.CLIENT.value

        log_info(InfoMessages.INITIALIZE_CLIENT.value, interface=interface)

        if not isinstance(config.vaultID, str):
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.VAULT_ID_INVALID_TYPE.value % (
                str(type(config.vaultID))), interface=interface)
        if not isinstance(config.vaultURL, str):
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.VAULT_URL_INVALID_TYPE.value % (
                str(type(config.vaultURL))), interface=interface)

        if not isinstance(config.tokenProvider, types.FunctionType):
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.TOKEN_PROVIDER_ERROR.value % (
                str(type(config.tokenProvider))), interface=interface)

        self.vaultID = config.vaultID
        self.vaultURL = config.vaultURL.rstrip('/')
        self.tokenProvider = config.tokenProvider
        self.storedToken = ''
        log_info(InfoMessages.CLIENT_INITIALIZED.value, interface=interface)

    def insert(self, records: dict, options: InsertOptions = InsertOptions()):
        interface = InterfaceName.INSERT.value
        log_info(InfoMessages.INSERT_TRIGGERED.value, interface=interface)

        self._checkConfig(interface)

        jsonBody = getInsertRequestBody(records, options)
        requestURL = self._get_complete_vault_url()
        self.storedToken = tokenProviderWrapper(
            self.storedToken, self.tokenProvider, interface)
        headers = {
            "Authorization": "Bearer " + self.storedToken
        }

        response = requests.post(requestURL, data=jsonBody, headers=headers)
        processedResponse = processResponse(response)
        result = convertResponse(records, processedResponse, options.tokens)

        log_info(InfoMessages.INSERT_DATA_SUCCESS.value, interface)
        return result

    def detokenize(self, records):
        interface = InterfaceName.DETOKENIZE.value
        log_info(InfoMessages.DETOKENIZE_TRIGGERED.value, interface)

        self._checkConfig(interface)
        self.storedToken = tokenProviderWrapper(
            self.storedToken, self.tokenProvider, interface)
        url = self._get_complete_vault_url() + '/detokenize'
        responses = asyncio.run(sendDetokenizeRequests(
            records, url, self.storedToken))
        result, partial = createDetokenizeResponseBody(responses)
        if partial:
            raise SkyflowError(SkyflowErrorCodes.PARTIAL_SUCCESS,
                               SkyflowErrorMessages.PARTIAL_SUCCESS, result, interface=interface)
        else:
            log_info(InfoMessages.DETOKENIZE_SUCCESS.value, interface)
            return result

    def get_by_id(self, records):
        interface = InterfaceName.GET_BY_ID.value
        log_info(InfoMessages.GET_BY_ID_TRIGGERED.value, interface)

        self._checkConfig(interface)
        self.storedToken = tokenProviderWrapper(
            self.storedToken, self.tokenProvider, interface)
        url = self._get_complete_vault_url()
        responses = asyncio.run(sendGetByIdRequests(
            records, url, self.storedToken))
        result, partial = createGetByIdResponseBody(responses)
        if partial:
            raise SkyflowError(SkyflowErrorCodes.PARTIAL_SUCCESS,
                               SkyflowErrorMessages.PARTIAL_SUCCESS, result, interface=interface)
        else:
            log_info(InfoMessages.GET_BY_ID_SUCCESS.value, interface)

            return result

    def invoke_connection(self, config: ConnectionConfig):

        interface = InterfaceName.INVOKE_CONNECTION.value
        log_info(InfoMessages.INVOKE_CONNECTION_TRIGGERED.value, interface)

        session = requests.Session()
        self.storedToken = tokenProviderWrapper(
            self.storedToken, self.tokenProvider, interface)
        request = createRequest(config)

        if not 'X-Skyflow-Authorization'.lower() in request.headers:
            request.headers['x-skyflow-authorization'] = self.storedToken

        response = session.send(request)
        session.close()
        return processResponse(response, interface=interface)

    def _checkConfig(self, interface):
        '''
            Performs basic check on the given client config
        '''
        if not len(self.vaultID) > 0:
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                               SkyflowErrorMessages.EMPTY_VAULT_ID, interface=interface)
        if not len(self.vaultURL) > 0:
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                               SkyflowErrorMessages.EMPTY_VAULT_URL, interface=interface)

    def _get_complete_vault_url(self):
        '''
            Get the complete vault url from given vault url and vault id
        '''
        return self.vaultURL + "/v1/vaults/" + self.vaultID



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from enum import Enum
from types import FunctionType
from typing import List


class Configuration:

    def __init__(self, vaultID: str=None, vaultURL: str=None, tokenProvider: FunctionType=None):
        
        self.vaultID = ''
        self.vaultURL = ''
          
        if tokenProvider == None and vaultURL == None and isinstance(vaultID, FunctionType):
            self.tokenProvider = vaultID
        elif tokenProvider == None and vaultID == None and isinstance(vaultURL, FunctionType):
            self.tokenProvider = vaultURL
        else:
            if tokenProvider is None:
                raise TypeError('tokenProvider must be given')
            self.vaultID = vaultID or ""
            self.vaultURL = vaultURL or ""
            self.tokenProvider = tokenProvider

class UpsertOption:
    def __init__(self,table: str,column: str):
        self.table = table
        self.column = column

class InsertOptions:
    def __init__(self, tokens: bool=True,upsert :List[UpsertOption]=None):
        self.tokens = tokens
        self.upsert = upsert

class RequestMethod(Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'

class ConnectionConfig:
    def __init__(self, connectionURL: str, methodName: RequestMethod, 
    pathParams: dict={}, queryParams: dict={}, requestHeader: dict={}, requestBody: dict={}):
        self.connectionURL = connectionURL.rstrip("/")
        self.methodName = methodName
        self.pathParams = pathParams
        self.queryParams = queryParams
        self.requestHeader = requestHeader
        self.requestBody = requestBody

class RedactionType(Enum):
    PLAIN_TEXT = "PLAIN_TEXT"
    MASKED = "MASKED"
    REDACTED = "REDACTED"
    DEFAULT = "DEFAULT"



'''
	Copyright (c) 2022 Skyflow, Inc.
'''
import json

import requests
from requests.models import HTTPError
from skyflow.errors._skyflow_errors import SkyflowError, SkyflowErrorCodes, SkyflowErrorMessages
from skyflow._utils import InterfaceName

interface = InterfaceName.INSERT.value


def getInsertRequestBody(data, options):
    try:
        records = data["records"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.RECORDS_KEY_ERROR, interface=interface)

    if not isinstance(records, list):
        recordsType = str(type(records))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_RECORDS_TYPE.value % (
            recordsType), interface=interface)
    
    upsertOptions = options.upsert 
    
    if upsertOptions:
        validateUpsertOptions(upsertOptions=upsertOptions)
            
    requestPayload = []
    insertTokenPayload = []
    for index, record in enumerate(records):
        tableName, fields = getTableAndFields(record)
        postPayload = {"tableName": tableName, "fields": fields,"method": "POST","quorum": True}
        
        if upsertOptions:
            postPayload["upsert"] = getUpsertColumn(tableName,upsertOptions)
        
        requestPayload.append(postPayload)
        if options.tokens:
            insertTokenPayload.append({
                "method": "GET",
                "tableName": tableName,
                "ID": "$responses." + str(index) + ".records.0.skyflow_id",
                "tokenization": True
            })
    requestBody = {"records": requestPayload + insertTokenPayload}
    try:
        jsonBody = json.dumps(requestBody)
    except Exception as e:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_JSON.value % (
            'insert payload'), interface=interface)

    return jsonBody


def getTableAndFields(record):
    try:
        table = record["table"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.TABLE_KEY_ERROR, interface=interface)

    if not isinstance(table, str):
        tableType = str(type(table))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_TABLE_TYPE.value % (
            tableType), interface=interface)

    try:
        fields = record["fields"]
    except KeyError:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT,
                           SkyflowErrorMessages.FIELDS_KEY_ERROR, interface=interface)

    if not isinstance(fields, dict):
        fieldsType = str(type(table))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_FIELDS_TYPE.value % (
            fieldsType), interface=interface)

    return (table, fields)


def processResponse(response: requests.Response, interface=interface):
    statusCode = response.status_code
    content = response.content.decode('utf-8')
    try:
        response.raise_for_status()
        try:
            return json.loads(content)
        except:
            raise SkyflowError(
                statusCode, SkyflowErrorMessages.RESPONSE_NOT_JSON.value % content, interface=interface)
    except HTTPError:
        message = SkyflowErrorMessages.API_ERROR.value % statusCode
        if response != None and response.content != None:
            try:
                errorResponse = json.loads(content)
                if 'error' in errorResponse and type(errorResponse['error']) == type({}) and 'message' in errorResponse['error']:
                    message = errorResponse['error']['message']
            except:
                message = SkyflowErrorMessages.RESPONSE_NOT_JSON.value % content
        if 'x-request-id' in response.headers:
            message += ' - request id: ' + response.headers['x-request-id']
        raise SkyflowError(statusCode, message, interface=interface)


def convertResponse(request: dict, response: dict, tokens: bool):
    responseArray = response['responses']
    records = request['records']
    recordsSize = len(records)
    result = []
    for idx, _ in enumerate(records):
        table = records[idx]['table']
        skyflow_id = responseArray[idx]['records'][0]['skyflow_id']
        if tokens:
            fieldsDict = responseArray[recordsSize + idx]['fields']
            fieldsDict['skyflow_id'] = skyflow_id
            result.append({'table': table, 'fields': fieldsDict})
        else:
            result.append({'table': table, 'skyflow_id': skyflow_id})
    return {'records': result}

def getUpsertColumn(tableName, upsertOptions):
    uniqueColumn:str = ''
    for upsertOption in upsertOptions:
        if tableName == upsertOption.table:
            uniqueColumn = upsertOption.column
    return uniqueColumn

def validateUpsertOptions(upsertOptions):
    if not isinstance(upsertOptions,list):
        upsertOptionsType = str(type(upsertOptions))
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_UPSERT_OPTIONS_TYPE.value %(
                upsertOptionsType),interface=interface)
    if len(upsertOptions) == 0:
        raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.EMPTY_UPSERT_OPTIONS_LIST.value, interface=interface)
    
    for index, upsertOption in enumerate(upsertOptions):
        if upsertOption.table == None or not isinstance(upsertOption.table,str):
            raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_UPSERT_TABLE_TYPE.value %(
                    index),interface=interface)
        if upsertOption.table == '':
                raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.EMPTY_UPSERT_OPTION_TABLE.value %(
                    index),interface=interface)
        if upsertOption.column == None or not isinstance(upsertOption.column,str):
                raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.INVALID_UPSERT_COLUMN_TYPE.value %(
                    index),interface=interface)
        if upsertOption.column == '':
                raise SkyflowError(SkyflowErrorCodes.INVALID_INPUT, SkyflowErrorMessages.EMPTY_UPSERT_OPTION_COLUMN.value %(
                    index),interface=interface)


'''
	Copyright (c) 2022 Skyflow, Inc.
'''
from setuptools import setup, find_packages
import sys

if sys.version_info < (3, 7):
    raise RuntimeError("skyflow requires Python 3.7+")
current_version = '1.7.0'

setup(
    name='skyflow',
    version=current_version,
    author='Skyflow',
    author_email='service-ops@skyflow.com',
    packages=find_packages(where='.', exclude=['test*']),
    url='https://github.com/skyflowapi/skyflow-python/',
    license='LICENSE',
    description='Skyflow SDK for the Python programming language',
    long_description=open('README.rst').read(),
    install_requires=[
        'PyJWT',
        'datetime',
        'requests',
        'aiohttp',
        'asyncio',
        'cryptography>=3.3.1'
    ],
    python_requires=">=3.7"
)
